# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" test polynomial module """

import numpy as np
import pytest

from quark import Polynomial
from quark.polynomial import sort_poly_dict, _get_info_values


X, Y = "x", "y"

VARIABLES = "variables"
OFFSET = "offset"
LINEAR = "linear"
QUADRATIC = "quadratic"

P1_DICT = {(1,): 5, (2,): 6, (3, 1): 3, (2, 4): 2, (4, 2): 1, (4,): 1, (): 3}
P2_DICT = {((Y, 2, 3), (X, 1)): 2, ((Y, 1, 2),): 3}
P3_DICT = {(2,): 1, (2, 4): 1, (3, 4): 3, (5,): 7}
P4_DICT = {((X, 1),): 7, ((X, 2),): 2}
P5_DICT = {(0,): 5, (1,): 6, (2, 0): -3, (1, 3): 0.5, (3,): -1, (0, 1): 0.9, (): 16, (2,): 0, (2, 2): 2}

POLY1 = Polynomial(P1_DICT)
POLY2 = Polynomial(P2_DICT)
POLY3 = Polynomial(P3_DICT)
POLY4 = Polynomial(P4_DICT)
POLY5 = Polynomial(P5_DICT)
POLY1_DUP = Polynomial(P1_DICT)
POLY2_DUP = Polynomial(P2_DICT)

POLY5_LINEAR = [ 5.,  6.,  0., -1.]
POLY5_QUADRATIC = [[ 0. ,  0.9, -3. ,  0. ],
                   [ 0. ,  0. ,  0. ,  0.5],
                   [ 0. ,  0. ,  2. ,  0. ],
                   [ 0. ,  0. ,  0. ,  0. ]]

# pylint: disable=protected-access


def teardown_function():
    """ reset cache """
    for poly in [POLY1, POLY2]:
        try:
            delattr(poly, VARIABLES)
            delattr(poly, LINEAR)
        except AttributeError:
            pass


def test_init():
    """ test initialization """
    assert len(POLY1) == 6
    assert len(POLY2) == 2
    assert (1, 3) in POLY1
    assert (3, 1) in POLY1
    assert (3, 1) not in dict(POLY1)
    assert ((X, 1), (Y, 2, 3)) in POLY2
    assert ((Y, 2, 3), (X, 1)) in POLY2
    assert ((Y, 2, 3), (X, 1)) not in dict(POLY2)
    assert all(isinstance(val, (float, int)) for val in POLY1.values())

    assert dict(POLY1) != P1_DICT
    assert dict(POLY2) != P2_DICT
    assert dict(POLY3) == P3_DICT
    assert dict(POLY4) == P4_DICT

    mixed_dict = {(1,): 5, ((Y, 1, 2),): 3}
    with pytest.raises(TypeError, match="Expected variable type 'Integral', but got 'tuple'"):
        Polynomial(mixed_dict)

    mixed_dict = {(1,): 5, ("a",): 6}
    with pytest.raises(TypeError, match="Expected variable type 'Integral', but got 'str'"):
        Polynomial(mixed_dict)

    mixed_dict = {("a",): 5, ((Y, 2, 3),): 6}
    with pytest.raises(TypeError, match="Expected variable type 'str', but got 'tuple'"):
        Polynomial(mixed_dict)

    strange_dict = {True : 5, False : 6}
    with pytest.raises(TypeError, match="Variable 'True' has invalid type 'bool'"):
        Polynomial(strange_dict)

    false_coeff_dict = {(1,) : "a"}
    with pytest.raises(ValueError, match="Invalid coefficient 'a'"):
        Polynomial(false_coeff_dict)

    false_tuple_dict = {((X, 1.4),) : 1}
    with pytest.raises(ValueError, match="is not formatted correctly, should only contain ints and strings"):
        Polynomial(false_tuple_dict)

def test_immutable():
    """ test disabled methods """
    with pytest.raises(TypeError, match="Cannot modify immutable Polynomial object"):
        POLY1[(1,)] = 4
    with pytest.raises(TypeError, match="Cannot modify immutable Polynomial object"):
        POLY2[(X, 1)] = 1
    with pytest.raises(TypeError, match="Cannot modify immutable Polynomial object"):
        del POLY1[(1,)]
    with pytest.raises(TypeError, match="Cannot modify immutable Polynomial object"):
        POLY1.pop((1,))
    with pytest.raises(TypeError, match="Cannot modify immutable Polynomial object"):
        POLY1.popitem((1,))
    with pytest.raises(TypeError, match="Cannot modify immutable Polynomial object"):
        POLY1.clear()
    with pytest.raises(TypeError, match="Cannot modify immutable Polynomial object"):
        POLY1.update({(1, 2) : 4})
    with pytest.raises(TypeError, match="Cannot modify immutable Polynomial object"):
        POLY1.setdefault((1,), 6)
    with pytest.raises(TypeError, match="Cannot modify immutable Polynomial object"):
        POLY1.setdefault((1, 2), 6)

def test_equal():
    """ test equality check """
    # pylint: disable=unneeded-not
    assert POLY1 != POLY3
    assert not POLY1 == POLY3
    assert POLY4 == POLY4.copy()
    assert POLY4 != POLY2

    # new initialization
    assert POLY1_DUP == POLY1
    assert not POLY1_DUP != POLY1
    assert not POLY1_DUP is POLY1
    assert POLY2 == POLY2_DUP
    assert not POLY2 != POLY2_DUP
    assert not POLY2 is POLY2_DUP

    # adding zero coefficient
    sum_poly = POLY1 + Polynomial({(0,): 0})
    assert sum_poly == POLY1
    assert not sum_poly != POLY1

    # constant
    assert Polynomial({() : 3}) == 3
    assert Polynomial({() : 4}) != 3
    assert Polynomial() + 3 == 3
    assert 3 + Polynomial() == 3
    assert Polynomial({() : 3, (1,) : 4}) != 3

    assert POLY3 != P3_DICT

def test_add():
    """ test addition """
    exp_poly = Polynomial({(): 3, (1,): 5, (2,): 7, (4,): 1, (5,): 7, (1, 3): 3, (2, 4): 4, (3, 4): 3})
    sum_poly = POLY1 + POLY3
    assert sum_poly == exp_poly
    assert POLY1 == POLY1_DUP

    sum_poly = POLY1
    sum_poly += POLY3
    assert sum_poly == exp_poly
    assert POLY1 == POLY1_DUP

    exp_poly = Polynomial({((Y, 2, 3), (X, 1)): 2, ((Y, 1, 2),): 3, () : 3})
    sum_poly = POLY2 + 3
    assert sum_poly == exp_poly

    sum_poly = POLY2
    sum_poly += 3
    assert sum_poly == exp_poly

    sum_poly = 3
    sum_poly += POLY2
    assert sum_poly == exp_poly

    with pytest.raises(TypeError, match="Can only add numbers or polynomials"):
        _ = POLY1 + (X, 1)
    with pytest.raises(TypeError, match="Expected variable type 'Integral', but got 'tuple'"):
        _ = POLY1 + POLY2

def test_sub():
    """ test subtraction """
    exp_poly = Polynomial({(): 3, (1,): 5, (2,): 5, (4,): 1, (5,): -7, (1, 3): 3, (2, 4): 2, (3, 4): -3})

    diff_poly = POLY1 - POLY3
    assert diff_poly == exp_poly
    assert POLY1 == POLY1_DUP

    diff_poly = POLY1
    diff_poly -= POLY3
    assert diff_poly == exp_poly
    assert POLY1 == POLY1_DUP

    exp_poly = Polynomial({((Y, 2, 3), (X, 1)): 2, ((Y, 1, 2),): 3, (): -3})
    diff_poly = POLY2 - 3
    assert diff_poly == exp_poly

    diff_poly = POLY2
    diff_poly -= 3
    assert diff_poly == exp_poly

    diff_poly = 3 - POLY2
    assert diff_poly == -1 * exp_poly

    diff_poly = 3
    diff_poly -= POLY2
    assert diff_poly == -1 * exp_poly

def test_mul():
    """ test multiplication """
    exp_poly = Polynomial({(2,): 3, (5,): 21,
                           (1, 2): 5, (1, 5): 35, (2, 2): 6, (2, 4): 4, (2, 5): 42, (3, 4): 9, (4, 5): 7,
                           (1, 2, 3): 3, (1, 2, 4): 5, (1, 3, 4): 15, (1, 3, 5): 21, (2, 2, 4): 9, (2, 3, 4): 18,
                           (2, 4, 4): 1, (2, 4, 5): 21, (3, 4, 4): 3,
                           (1, 2, 3, 4): 3, (1, 3, 3, 4): 9, (2, 2, 4, 4): 3, (2, 3, 4, 4): 9})

    prod_poly = POLY1 * POLY3
    assert prod_poly == exp_poly
    assert POLY1 == POLY1_DUP

    prod_poly = POLY1
    prod_poly *= POLY3
    assert prod_poly == exp_poly
    assert POLY1 == POLY1_DUP

    prod_poly = POLY1
    prod_poly *= POLY1
    assert POLY1 * POLY1 == prod_poly

    exp_poly = Polynomial({((Y, 2, 3), (X, 1)): 4, ((Y, 1, 2),): 6})
    prod_poly = POLY2 * 2
    assert prod_poly == exp_poly
    assert POLY2 == POLY2_DUP

    prod_poly = POLY2
    prod_poly *= 2
    assert prod_poly == exp_poly
    assert POLY2 == POLY2_DUP

    prod_poly = 2
    prod_poly *= POLY2
    assert prod_poly == exp_poly
    assert POLY2 == POLY2_DUP

    with pytest.raises(TypeError, match="Can only multiply with numbers or polynomials"):
        _ = POLY1 * (X, 1)
    with pytest.raises(TypeError, match="Expected variable type 'Integral', but got 'str'"):
        _ = POLY1 * Polynomial({X : 1})

    poly = Polynomial({((X, 1),): 2, ((X, "a"),): 1})
    with pytest.raises(ValueError, match="Sorting failed"):
        _ = poly * poly

def test_div():
    """ test division """
    poly = Polynomial({((Y, 2, 3), (X, 1)): 4, ((Y, 1, 2),): 6})
    div_poly = poly / 2
    assert div_poly == POLY2

    div_poly = poly
    div_poly /= 2
    assert div_poly == POLY2

    div_poly = 2
    with pytest.raises(TypeError, match=r"unsupported operand type\(s\) for \/=: 'int' and 'Polynomial'"):
        div_poly /= poly
    with pytest.raises(TypeError, match="Divisor should be number"):
        _ = poly / POLY1

def test_pow():
    """ test potentiated polynomials """
    exp_poly = Polynomial({((X, 1), (X, 1)): 49, ((X, 2), (X, 2)): 4, ((X, 1), (X, 2)): 28})
    assert POLY4 ** 2 == exp_poly
    assert POLY4 ** 3 == exp_poly * POLY4

    with pytest.raises(TypeError, match="Exponent should be integer"):
        _ = POLY1 ** 1.4

def test_contains():
    """ test containment of monomials """
    assert (2, 4) in POLY1
    assert (4, 2) in POLY1
    assert ((Y, 2, 3), (X, 1)) in POLY2
    assert ((X, 1), (Y, 2, 3)) in POLY2

def test_get():
    """ test getting of coefficients of monomials """
    assert POLY1.get((2, 4)) == 3
    assert POLY1.get((4, 2)) == 3
    assert POLY2.get(((Y, 2, 3), (X, 1))) == 2
    assert POLY2.get(((X, 1), (Y, 2, 3))) == 2

def test_getitem():
    """ test getting of coefficients using items """
    assert POLY1[2, 4] == 3
    assert POLY1[4, 2] == 3
    assert POLY2[(Y, 2, 3), (X, 1)] == 2
    assert POLY2[(X, 1), (Y, 2, 3)] == 2

def test_repr():
    """ test string representation """
    assert repr(POLY1) == "Polynomial({(): 3, (1,): 5, (2,): 6, (4,): 1, (1, 3): 3, (2, 4): 3})"

def test_str():
    """ test string representation """
    assert str(POLY1) == "+3 +5 x1 +6 x2 +1 x4 +3 x1 x3 +3 x2 x4"

def test_copy():
    """ test copying """
    copied_poly = POLY1.copy()
    assert POLY1 == copied_poly
    assert POLY1 is not copied_poly

def test_degree():
    """ test degree """
    assert POLY3.degree == 2
    assert POLY4.degree == 1
    assert (POLY4 * POLY4).degree == 2

def test_is_quadratic():
    """ test check for degree at most 2 """
    assert POLY1.is_quadratic()
    assert POLY4.is_quadratic()

    prod_poly = POLY3 * POLY3
    assert not prod_poly.is_quadratic()

    prod_poly = POLY4 * POLY4
    assert prod_poly.is_quadratic()

def test_has_int_coefficients():
    """ check for integer polynomials """
    assert POLY1.has_int_coefficients()
    assert POLY4.has_int_coefficients()
    assert Polynomial({(1,): 1}).has_int_coefficients()
    assert not Polynomial({(1,): 1.01}).has_int_coefficients()

def test_is_flat():
    """ check for flat variables """
    assert POLY1.is_flat()
    assert not POLY2.is_flat()
    assert POLY3.is_flat()
    assert not POLY4.is_flat()
    assert POLY5.is_flat()

def test_is_compact():
    """ test check for compact variables """
    assert POLY5.is_compact()
    assert not POLY1.is_compact()
    assert not POLY2.is_compact()

    poly = Polynomial({(0,) : 1})
    assert poly.is_compact()
    assert (POLY1 + poly).is_compact()

    compact_poly = POLY1.compact()
    assert compact_poly.is_compact()
    compact_poly = POLY2.compact()
    assert compact_poly.is_compact()

    assert Polynomial().is_compact()
    assert Polynomial({() : 3}).is_compact()

def test_variables():
    """ test list of variables """
    exp_variables = [1, 2, 3, 4]
    assert VARIABLES not in POLY1.__dict__
    assert POLY1.variables == exp_variables
    assert VARIABLES in POLY1.__dict__
    assert POLY1.variables == exp_variables

    exp_variables = [(X, 1), (Y, 1, 2), (Y, 2, 3)]
    assert VARIABLES not in POLY2.__dict__
    assert POLY2.variables == exp_variables
    assert VARIABLES in POLY2.__dict__
    assert POLY2.variables == exp_variables

def test_offset():
    """ test constant offset """
    exp_offset = 3
    assert OFFSET not in POLY1.__dict__
    assert POLY1.offset == exp_offset

    exp_offset = 0
    assert OFFSET not in POLY2.__dict__
    assert POLY2.offset == exp_offset

def test_linear():
    """ test linear part """
    exp_linear = Polynomial({(1,): 5, (2,): 6, (3,): 0, (4,): 1})
    assert LINEAR not in POLY1.__dict__
    assert POLY1.linear == exp_linear

    exp_linear = Polynomial({((X, 1),): 0, ((Y, 1, 2),): 3, ((Y, 2, 3),): 0})
    assert LINEAR not in POLY2.__dict__
    assert POLY2.linear == exp_linear

def test_linear_plain():
    """ test linear part """
    exp_linear = {1: 5, 2: 6, 3: 0, 4: 1}
    assert LINEAR not in POLY1.__dict__
    assert POLY1.linear_plain == exp_linear

    exp_linear = {(X, 1): 0, (Y, 1, 2): 3, (Y, 2, 3): 0}
    assert LINEAR not in POLY2.__dict__
    assert POLY2.linear_plain == exp_linear

def test_quadratic():
    """ test quadratic part """
    exp_quadratic = Polynomial({(1, 3): 3, (2, 4): 3})
    assert QUADRATIC not in POLY1.__dict__
    assert POLY1.quadratic == exp_quadratic

    exp_quadratic = Polynomial({((X, 1), (Y, 2, 3)): 2})
    assert QUADRATIC not in POLY2.__dict__
    assert POLY2.quadratic == exp_quadratic

def test_get_variable_num():
    """ test number of variables """
    assert VARIABLES not in POLY1.__dict__
    assert POLY1.get_variable_num() == 4
    assert VARIABLES in POLY1.__dict__

    assert VARIABLES not in POLY2.__dict__
    assert POLY2.get_variable_num() == 3
    assert VARIABLES in POLY2.__dict__

def test_get_coefficient():
    """ test coefficients """
    assert POLY1.get_coefficient(1) == 5
    assert POLY1.get_coefficient(3) == 0

    assert POLY2.get_coefficient((Y, 1, 2)) == 3
    assert POLY2.get_coefficient((X, 2)) == 0

    assert POLY1.get_coefficient(1, 3) == 3
    assert POLY1.get_coefficient(3, 1) == 3
    assert POLY1.get_coefficient(1, 6) == 0

    assert POLY2.get_coefficient((X, 1), (Y, 2, 3)) == 2
    assert POLY2.get_coefficient((X, 2), (Y, 2, 3)) == 0

def test_sigmas():
    """ test sigma values """
    exp_sigmas = {1: (3, 0), 2: (3, 0), 3: (3, 0), 4: (3, 0)}
    assert POLY1.sigmas == exp_sigmas

    exp_sigmas = {(X, 1): (2, 0), (Y, 2, 3): (2, 0), (Y, 1, 2): (0, 0)}
    assert POLY2.sigmas == exp_sigmas

    exp_sigmas = {0: (0.9, -3), 1: (1.4, 0), 2: (4, -3), 3: (0.5, 0)}
    assert POLY5.sigmas == exp_sigmas

def test_remove_zero_coefficients():
    """ test removal of monomials with zero coefficient """
    poly1_with_zeros = Polynomial({(1,): 5, (2,): 6, (3,): 0, (3, 1): 3, (2, 4): 3, (4,): 1, (): 3, (1, 4): 0})
    assert dict(poly1_with_zeros) != dict(POLY1)
    assert poly1_with_zeros == POLY1

    poly1_without_zeros = poly1_with_zeros.remove_zero_coefficients()
    assert dict(poly1_without_zeros) == dict(POLY1)
    assert poly1_without_zeros == POLY1

def test_evaluate():
    """ test evaluation """
    # P1(x0, x1, x2, x3, x4) = 5 x1 + 6 x2 + 3 x1 x3 + x4 + 3 x2 x4 + 3
    # P1(0, 1, 0, 1, 1) = 5 + 3 + 1 + 3 = 12
    assert POLY1.evaluate([0, 1, 0, 1, 1]) == 12

    # P1(-1, 1, -1, 1, 1) = 5 - 6 + 3 + 1 - 3 + 3 = 3
    assert POLY1.evaluate([-1, 1, -1, 1, 1]) == 3

    # P1(-1, 1, -1, 1, 1) = 5 - 6 + 3 + 1 - 3 + 3 = 3
    assert POLY1.evaluate([-1, 1, -1, 1, 1], keep_poly=True) == Polynomial({(): 3})

    exp_poly = Polynomial({(): 2, (3,): 3, (4,): -2})
    assert POLY1.evaluate([-1, 1, -1]) == exp_poly

    var_assignment = [Polynomial({(i,) : -1}) for i in range(5)]
    exp_poly = Polynomial({(1,): -5, (2,): -6, (3, 1): 3, (2, 4): 3, (4,): -1, (): 3})
    assert POLY1.evaluate(var_assignment) == exp_poly

    var_assignment = [1, 1, Polynomial({(2,) : -1}), 0, -2]
    assert POLY1.evaluate(var_assignment) == 6

    exp_poly = Polynomial({(2,): 6, (3,): 3, (2, 4): 3, (4,): 1, (): 8})
    assert POLY1.evaluate({1: 1}) == exp_poly

    exp_poly = Polynomial({(2,): 6, (2, 4): 3, (4,): 1, (): 3})
    assert POLY1.evaluate({1: 0}) == exp_poly

    exp_poly = Polynomial({(): 21})
    assert POLY1.evaluate({v: 1 for v in [1, 2, 3, 4]}) == exp_poly

    with pytest.raises(ValueError, match="Kept variables and replaced variables are not compatible"):
        POLY3.evaluate({4: Polynomial({((X, 1),): 1})})

def test_affine_transformation():
    """ test affine transformation """
    transformed_poly = POLY1.affine_transform(3, 5)
    exp_poly = Polynomial.read_from_string("60 x1 + 63 x2 + 27 x1 x3 + 45 x3 + 48 x4 + 27 x2 x4 + 213")
    assert transformed_poly == exp_poly

def test_replace_variables():
    """ test replacement of variables """
    replacement = {1 : "b", 2 : "c", 3 : "d", 4 : "e"}
    exp_poly = Polynomial({("b",): 5, ("c",): 6, ("d", "b"): 3, ("c", "e"): 3, ("e",): 1, (): 3})
    assert POLY1.replace_variables(replacement) == exp_poly

    replacement = ["a", "b", "c", "d", "e"]
    exp_poly = Polynomial({("b",): 5, ("c",): 6, ("d", "b"): 3, ("c", "e"): 3, ("e",): 1, (): 3})
    assert POLY1.replace_variables(replacement) == exp_poly

    replacement = {(Y, 2, 3) : 0, (X, 1) : 1, (Y, 1, 2) : 2}
    exp_poly = Polynomial({(0, 1): 2, (2,): 3})
    assert POLY2.replace_variables(replacement) == exp_poly

    # only partially replacing with variables of a different type does not work
    # in different monomials
    replacement = {(Y, 2, 3): "a", (Y, 1, 2): "b"}
    with pytest.raises(ValueError, match="Replacement is invalid"):
        POLY2.replace_variables(replacement)
    # in single monomial
    replacement = {(Y, 2, 3) : "a", (X, 1) : "b"}
    with pytest.raises(ValueError, match="Replacement is invalid"):
        POLY2.replace_variables(replacement)

    replacement = {(Y, 2, 3) : ("z", 1), (X, 1) : ("z", 2)}
    exp_poly = Polynomial({(("z", 1), ("z", 2)): 2, ((Y, 1, 2),): 3})
    assert POLY2.replace_variables(replacement) == exp_poly

    replacement = {var : 0 for var in POLY1.variables}
    exp_poly = Polynomial({(0,): 12, (0, 0): 6, (): 3})
    assert POLY1.replace_variables(replacement) == exp_poly

    replacement = [0] * 5
    exp_poly = Polynomial({(0,): 12, (0, 0): 6, (): 3})
    assert POLY1.replace_variables(replacement) == exp_poly

    exp_poly = Polynomial({((X, 1),): 5, ((X, 2),): 6, ((X, 3), (X, 1)): 3, ((X, 2), (X, 4)): 2,
                             ((X, 4), (X, 2)): 1, ((X, 4),): 1, (): 3})
    assert POLY1.replace_variables(lambda x: (X, x)) == exp_poly

def test_replace_variables_by_ordering():
    """ test replacement of variables """
    replacement = [(Y, 2, 3), (X, 1), (Y, 1, 2)]
    exp_poly = Polynomial({(0, 1): 2, (2,): 3})
    assert POLY2.replace_variables_by_ordering(replacement) == exp_poly

def test_compact():
    """ test construction of compact polynomials """
    compact_poly = POLY2.compact()
    exp_poly = Polynomial({(0, 2): 2, (1,): 3})
    assert compact_poly == exp_poly
    assert compact_poly.variables == [0, 1, 2]

def test_round():
    """ test rounding """
    exp_poly = Polynomial({(0,): 5, (1,): 6, (2, 0): -3, (2, 2): 2, (1, 3): 0, (3,): -1, (0, 1): 1, (): 16})
    assert POLY5.round(0) == exp_poly

    exp_poly = Polynomial({(1,): 4, (2,): 5.78})
    assert Polynomial({(1,): 4, (2,): 5.7777}).round(2) == exp_poly

def test_get_coeff_list():
    """ test coefficients lists """
    assert len(POLY5.coefficients_lists) == 3
    assert POLY5.coefficients_lists[0] == [16]
    assert POLY5.coefficients_lists[1] == [-1, 0, 5, 6]
    assert POLY5.coefficients_lists[2] == [-3, 0.5, 0.9, 2.0]

def test_get_coefficients_info_by_degree():
    """ test coefficient calculation """
    exp_info = {"max_abs_coeff_degree_1":      6.0,
                "min_abs_coeff_degree_1":      1.0,
                "min_dist_degree_1":           1.0,
                "min_abs_dist_degree_1":       1.0,
                "max_min_ratio_degree_1":      6.0,
                "max_dist_ratio_degree_1":     6.0,
                "max_abs_dist_ratio_degree_1": 6.0}
    assert POLY5.get_coefficients_info_by_degree(1) == exp_info

    exp_info = {"max_abs_coeff_degree_2":      3.0,
                "min_abs_coeff_degree_2":      0.5,
                "min_dist_degree_2":           0.4,
                "min_abs_dist_degree_2":       0.4,
                "max_min_ratio_degree_2":      6.0,
                "max_dist_ratio_degree_2":     7.5,
                "max_abs_dist_ratio_degree_2": 7.5}
    assert POLY5.get_coefficients_info_by_degree(2) == exp_info

    exp_info = {"max_abs_coeff_degree_1":      7.0,
                "min_abs_coeff_degree_1":      2.0,
                "min_dist_degree_1":           2.0,
                "min_abs_dist_degree_1":       2.0,
                "max_min_ratio_degree_1":      3.5,
                "max_dist_ratio_degree_1":     3.5,
                "max_abs_dist_ratio_degree_1": 3.5,}
    assert POLY4.get_coefficients_info_by_degree(1) == exp_info
    assert not POLY4.get_coefficients_info_by_degree(2)

    with pytest.raises(ValueError, match="Provide degree greater than 0"):
        POLY4.get_coefficients_info_by_degree(0)

def test_get_all_coefficients_info():
    """ test coefficient calculation """
    exp_info = {"max_abs_coeff_degree_1":      6.0, "max_abs_coeff_degree_2":      3.0, "max_abs_coeff":      6.0,
                "min_abs_coeff_degree_1":      1.0, "min_abs_coeff_degree_2":      0.5, "min_abs_coeff":      0.5,
                "min_dist_degree_1":           1.0, "min_dist_degree_2":           0.4, "min_dist":           0.4,
                "min_abs_dist_degree_1":       1.0, "min_abs_dist_degree_2":       0.4, "min_abs_dist":       0.1,
                "max_min_ratio_degree_1":      6.0, "max_min_ratio_degree_2":      6.0, "max_min_ratio":      12.0,
                "max_dist_ratio_degree_1":     6.0, "max_dist_ratio_degree_2":     7.5, "max_dist_ratio":     15.0,
                "max_abs_dist_ratio_degree_1": 6.0, "max_abs_dist_ratio_degree_2": 7.5, "max_abs_dist_ratio": 60.0}
    assert POLY5.get_all_coefficients_info() == exp_info

    exp_info = {"max_abs_coeff_degree_1":      7.0, "max_abs_coeff":      7.0,
                "min_abs_coeff_degree_1":      2.0, "min_abs_coeff":      2.0,
                "min_dist_degree_1":           2.0, "min_dist":           2.0,
                "min_abs_dist_degree_1":       2.0, "min_abs_dist":       2.0,
                "max_min_ratio_degree_1":      3.5, "max_min_ratio":      3.5,
                "max_dist_ratio_degree_1":     3.5, "max_dist_ratio":     3.5,
                "max_abs_dist_ratio_degree_1": 3.5, "max_abs_dist_ratio": 3.5}
    assert POLY4.get_all_coefficients_info() == exp_info

    assert not Polynomial().get_all_coefficients_info()

def test_calc_coefficient_info_from_list():
    """ test coefficient calculation """
    coefficients_list = [10, -10, 8, 7, 0.1, 1.1, 1.15, 13, -1.12]
    exp_values = (0.1, 13, 0.05, 0.02, 13 / 0.1, 13 / 0.05, 13 / 0.02)
    values = _get_info_values(coefficients_list)
    assert values == pytest.approx(exp_values, abs=1e-9)
    _check_coefficient_info_calc(coefficients_list)
    assert _get_info_values([]) == (0, 0, 0, 0, 0, 0, 0)

def test_calc_coefficient_info_from_list_random():
    """ test coefficient calculation """
    np.random.seed(0)
    _check_coefficient_info_calc(list(np.random.randn(20)))
    _check_coefficient_info_calc(list(np.random.uniform(-10, 0, 20)))
    _check_coefficient_info_calc(list(np.random.uniform(0, 10, 20)))

def test_to_string():
    """ test string creation """
    assert str(POLY1) == "+3 +5 x1 +6 x2 +1 x4 +3 x1 x3 +3 x2 x4"
    assert str(POLY2) == "+3 y_1_2 +2 x_1 y_2_3"
    assert str(POLY3) == "+1 x2 +7 x5 +1 x2 x4 +3 x3 x4"
    assert str(Polynomial({X : 4, (X, Y) : 2})) == "+4 x +2 x y"
    assert str(Polynomial({((1, 1),) : 4, ((1, 2), (0, 2)) : 2})) == "+4 x_1_1 +2 x_0_2 x_1_2"

    exp_str = "+3 +4 x0 +6 x1 +3 x2 +8 x0 x1 +6 x1 x2 +3 x1 x3 +4 x0 x1 x3 +3 x1 x2 x3"
    assert str(Polynomial({1 : 2, (3, 1) : 1, () : 1}) * Polynomial({2 : 3, 0 : 4, () : 3})) == exp_str

def test_sorted_dict():
    """ test sorting of poly dictionary """
    poly = Polynomial({(3,): 2, (1, 2): 4, (1,): 1})
    assert str(dict(poly)) == "{(3,): 2, (1, 2): 4, (1,): 1}"
    assert str(poly) == "+1 x1 +2 x3 +4 x1 x2"
    assert str(dict(poly)) == "{(1,): 1, (3,): 2, (1, 2): 4}"

    sorted_poly_dict = sort_poly_dict(poly)
    assert list(sorted_poly_dict.keys()) == [(1,), (3,), (1, 2)]
    sorted_poly_dict = sort_poly_dict(poly, reverse_monomial_degree_order=True)
    assert list(sorted_poly_dict.keys()) == [(1, 2), (1,), (3,)]

    poly = Polynomial({((X, 1),): 2, ((X, Y),): 1})
    with pytest.raises(ValueError, match="Sorting failed"):
        str(poly)

def test_read_from_string():
    """ test reading in from strings """
    exp_poly = Polynomial({(2,): 3, (5,): 21,
                           (1, 5): 35, (1, 12): -5, (2, 2): 6, (2, 4): 4, (2, 5): 42, (3, 4): 9, (5, 41): -7,
                           (1, 2, 3): 3, (1, 2, 4): 5, (1, 3, 4): 15, (1, 3, 5): 21, (2, 2, 4): 9, (2, 3, 4): 18,
                           (2, 4, 4): 1, (3, 4, 4): 3,
                           (1, 2, 3, 4): 3, (1, 3, 3, 4): -9, (2, 2, 4, 4): 3, (2, 3, 4, 4): 9, (2, 4, 5, 7): 21})

    input_str = """3.0 x2 - 5 x1 x12 + 6. x2^2 + 3 x1 x2 x3 + 4 x2 x4 + 5 x1 x2 x4 + 9 x2^2 x4 + 9 x3 x4 + 15 x1 x3 x4
                   + 18 x2 x3 x4 + 3 x1 x2 x3 x4 - 9 x1 x3^2 x4 + 1 x2 x4^2 + 3 x2^2 x4^2 + 3 x3 x4^2 + 9 x2 x3 x4^2
                   + 21 x5 + 35 x1 x5 + 42 x2 x5 + 21 x1 x3 x5 - 7 x41 x5 + 21 x2 x4 x5 x7"""
    poly_from_str = Polynomial.read_from_string(input_str)
    assert poly_from_str == exp_poly

    output_str = str(poly_from_str)
    assert Polynomial.read_from_string(output_str) == poly_from_str

    with pytest.raises(ValueError, match="Double exponents are not supported"):
        Polynomial.read_from_string("1. x1^2^3")

def _check_coefficient_info_calc(coeff_list):
    exp_min_abs_coeff = min(abs(v) for v in coeff_list)
    exp_max_abs_coeff = max(abs(v) for v in coeff_list)
    exp_min_dist = min(abs(v1 - v2) for v1 in coeff_list for v2 in coeff_list if abs(v1 - v2) != 0)
    exp_min_abs_dist = min(abs(abs(v1) - abs(v2)) for v1 in coeff_list for v2 in coeff_list
                                 if abs(abs(v1) - abs(v2)) != 0)
    exp_max_min_ratio = exp_max_abs_coeff / exp_min_abs_coeff
    exp_max_dist_ratio = exp_max_abs_coeff / exp_min_dist
    exp_max_abs_dist_ratio = exp_max_abs_coeff / exp_min_abs_dist

    exp_coeff_info = (exp_min_abs_coeff, exp_max_abs_coeff, exp_min_dist, exp_min_abs_dist, exp_max_min_ratio,
                      exp_max_dist_ratio, exp_max_abs_dist_ratio)
    assert _get_info_values(coeff_list) == pytest.approx(exp_coeff_info, abs=1e-9)

def test_get_matrix_representation():
    """ test standard matrix representation """
    with pytest.warns(UserWarning, match="Constant offset of 16 is dropped"):
        linear, quadratic = POLY5.get_matrix_representation()
    assert np.array_equal(linear, POLY5_LINEAR)
    assert np.array_equal(quadratic, POLY5_QUADRATIC)

    poly1_linear = [0, 5, 6, 0, 1, 0]
    poly1_quadratic = [[0, 0, 0, 0, 0, 0],
                       [0, 0, 0, 3, 0, 0],
                       [0, 0, 0, 0, 3, 0],
                       [0, 0, 0, 0, 0, 0],
                       [0, 0, 0, 0, 0, 0],
                       [0, 0, 0, 0, 0, 0]]
    with pytest.warns(UserWarning, match="Constant offset of 3 is dropped"):
        linear, quadratic = POLY1.get_matrix_representation(num_variables=6)
    assert np.array_equal(linear, poly1_linear)
    assert np.array_equal(quadratic, poly1_quadratic)

    with pytest.raises(ValueError, match="Only applicable for flat polynomials"):
        POLY2.get_matrix_representation()
    with pytest.raises(ValueError, match="Either provide compact polynomial or the total number of variables"):
        POLY1.get_matrix_representation()
    with pytest.raises(ValueError, match="Only applicable for quadratic polynomials"):
        Polynomial({(1, 2, 3) : 1}).get_matrix_representation()

def test_get_from_matrix_representation():
    """ test conversion from matrix representation """
    poly = Polynomial.get_from_matrix_representation(POLY5_LINEAR, POLY5_QUADRATIC)
    assert poly == POLY5 - 16

    with pytest.raises(ValueError, match="Can only process matrices with at most two dimensions"):
        Polynomial.get_from_matrix_representation([[[1]]])

def test_get_graph():
    """ test graph generation """
    graph = POLY1.get_graph()
    exp_nodes = {1, 2, 3, 4}
    exp_edges = {(1, 3), (2, 4)}
    assert set(graph.nodes) == exp_nodes
    assert set(graph.edges) == exp_edges

    with pytest.raises(ValueError, match="Only applicable for quadratic polynomials"):
        (POLY1 * POLY1).get_graph()



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for ScipModel """

from warnings import warn
from ast import literal_eval
from numbers import Real
from pyscipopt.scip import Model  # pylint: disable=no-name-in-module

from .solution import Solution


VAR_PREFIX                 = "var_"
OBJECTIVE_VAR              = "obj"
FILENAME_SUFFIX            = ".cip"
QUADRATIC_OBJECTIVE        = "quadratic_objective"
FROM_CONSTRAINED_OBJECTIVE = "from_constrained_objective"
FROM_OBJECTIVE             = "from_objective"
LIMITS_TIME                = "limits/time"
BINARY                     = "B"
CONTINUOUS                 = "C"
SAVE_ISING                 = "# from_inverted_ising: "

INFEASIBLE = "infeasible"
OPTIMAL    = "optimal"
UNBOUNDED  = "unbounded"
TIMEOUT    = "timelimit"
IS_SOLVED = [INFEASIBLE, OPTIMAL, UNBOUNDED]
HAS_SOLUTION = [OPTIMAL, TIMEOUT]

ERROR_VALUE  = "Cannot get value for variable '{}'"
WARNING_LOAD = "Cannot retrieve value for 'from_inverted_ising', will use 'None'"


class ScipModel(Model):
    """
    A Scip model provides an interface to the classical MIP solver SCIP.
    It automatically builds up the SCIP model from the ConstrainedObjective or the Objective,
    which can be solved with SCIP.
    """

    def __init__(self, name="", from_inverted_ising=None):
        """
        initialize ScipModel object

        :param (str) name: the name of the model
        :param (bool or None) from_inverted_ising: if None, originates from a binary problem,
                                                   if False, originates from a non-inverted (default) Ising model,
                                                   if True, originates from an inverted Ising model
        """
        self.from_inverted_ising = from_inverted_ising
        self.data = {}
        super().__init__(name)
        self.hideOutput()

    @property
    def name(self):
        """ the name of the model """
        return self.getProbName()

    @classmethod
    def get_from_objective(cls, objective):
        """
        get the corresponding ScipModel to an at most quadratic Objective

        :param (Objective) objective: the Objective with the quadratic Ising or binary polynomial
        :return: the corresponding ScipModel
        """
        name = objective.name or FROM_OBJECTIVE
        model = cls(name) if objective.is_binary() else cls(name, objective.is_ising(inverted=True))
        model._init_from_objective(objective)
        return model

    def _init_from_objective(self, objective):
        """ initialize the ScipModel with the data from the Objective """
        self._init_from_binary_poly(objective.to_binary().polynomial)

    def _init_from_binary_poly(self, binary_poly):
        """ initialize the ScipModel with the data from the binary polynomial corresponding to the Objective """
        # initialize binary variables (do not need to be flat)
        variables = {var: self.addVar(VAR_PREFIX + f"{var}", vtype=BINARY) for var in binary_poly.variables}
        # initialize a continuous objective variable serving as a placeholder
        objective = self.addVar(OBJECTIVE_VAR, vtype=CONTINUOUS, lb=-self.infinity())
        # get quadratic scip expression
        objective_scip_expression = binary_poly.evaluate(variables)
        # as scip cannot have quadratic objective functions directly, we set it equal to another continuous variable
        self.addCons(objective_scip_expression == objective, name=QUADRATIC_OBJECTIVE)
        # which we can then optimize as it is linear
        self.setObjective(objective)
        self.setMinimize()
        # container for some data, we store the mapping of the variables
        self.data = variables

    @classmethod
    def get_from_constrained_objective(cls, constrained_objective, name=FROM_CONSTRAINED_OBJECTIVE):
        """
        get the corresponding ScipModel to a ConstrainedObjective with an at most quadratic objective function

        :param (ConstrainedObjective) constrained_objective: the ConstrainedObjective with the quadratic Ising or
                                                             binary objective polynomial
        :param (str) name: the name of the resulting model, by default 'from_constrained_objective'
        :return: the corresponding ScipModel
        """
        model = cls(name)
        model._init_from_constrained_objective(constrained_objective)
        return model

    def _init_from_constrained_objective(self, constrained_objective):
        """ initialize the ScipModel with the data from the ConstrainedObjective """
        # initialize binary variables (do not need to be flat)
        variables = {var: self.addVar(VAR_PREFIX + f"{var}", vtype=BINARY)
                     for var in constrained_objective.get_all_variables()}

        objective_scip_expression = constrained_objective.objective_poly.evaluate(variables)
        if constrained_objective.objective_poly.is_linear():
            # if the objective function is linear we can directly use it in SCIP
            self.setObjective(objective_scip_expression)
        else:
            # otherwise we need to use a placeholder variable as the objective function and add a constraint
            objective = self.addVar(OBJECTIVE_VAR, vtype=CONTINUOUS, lb=-self.infinity())
            self.addCons(objective_scip_expression == objective, name=QUADRATIC_OBJECTIVE)
            self.setObjective(objective)

        # also add the constraints
        for name, constraint in constrained_objective.items():
            # since SCIP does not allow constants in the constraint function, we need to extract it
            shifted_constraint = constraint - constraint.polynomial.offset
            shifted_constraint.polynomial = shifted_constraint.polynomial.remove_zero_coefficients()
            poly_scip_expression = shifted_constraint.polynomial.evaluate(variables)
            # the brackets are necessary for SCIP to have both constraints (left and right side) added at the same time
            self.addCons(float(shifted_constraint.lower_bound) <=
                         (poly_scip_expression <= float(shifted_constraint.upper_bound)), name=name)

        self.setMinimize()
        # store variable mapping in data container
        self.data = variables

    def solve(self, timeout=3600.0, verbose=False, **param_kwargs):
        """
        solve the model using SCIP with the given parameters

        :param (float or int) timeout: the time in seconds until which SCIP tries to find the solution
        :param (bool) verbose: hide or print the output of SCIP
        :param param_kwargs: further parameters which can be given to SCIP,
                             see https://www.scipopt.org/doc-7.0.1/html/PARAMETERS.php
        :return: the optimal solution or best solution at timeout
        """
        self.set_solving_parameters(timeout, verbose, **param_kwargs)
        self.optimize()
        self.hideOutput()
        return self.get_solution()

    def set_solving_parameters(self, timeout=3600.0, verbose=False, **param_kwargs):
        """
        set the solving parameters to the model

        :param (float or int) timeout: the time in seconds until which SCIP tries to find the solution
        :param (bool) verbose: hide or print the output of SCIP
        :param param_kwargs: further parameters which can be given to SCIP,
                             see https://www.scipopt.org/doc-7.0.1/html/PARAMETERS.php
        """
        self.hideOutput(not verbose)
        self.setParam(LIMITS_TIME, timeout)
        for param, value in param_kwargs.items():
            self.setParam(param, value)

    def get_solution(self):
        """
        get the solution from the model, can only be applied once the model is solved,
        the type of the solution corresponds to the type of the original objective
        """
        objective_value, var_assignment = self.get_values()
        solution = Solution(var_assignments=var_assignment,
                            objective_value=objective_value,
                            solving_status=self.getStatus(),
                            solving_success=self.getStatus() in IS_SOLVED,
                            solving_time=self.getSolvingTime(),
                            total_time=self.getTotalTime(),
                            dual_gap=self.getGap(),
                            dual_bound=self.getDualbound(),
                            name=self.name)
        if self.from_inverted_ising is not None:
            solution = solution.to_ising(invert=self.from_inverted_ising)
        return solution

    def get_values(self):
        """ get the objective value and the values for each variable """
        if self.getStatus() in HAS_SOLUTION:
            objective_value = self.getObjVal()
            var_assignment = {var: self.get_value(var) for var in self.data}
            return objective_value, var_assignment
        return None, {}

    def get_value(self, variable, int_precision=1e-9):
        """
        get the assignment of the variable in the solution

        :param (tuple or Integral or str) variable: the original variable
        :param (Real) int_precision: precision value for checking if the value is actually an integer
        :return: the value of the variable in the solution
        """
        try:
            scip_var = self.data[variable]
            value = scip_var if isinstance(scip_var, Real) else self.getVal(scip_var)
        except KeyError as e:
            raise ValueError(ERROR_VALUE.format(variable)) from e
        value = round(value) if abs(value - round(value)) < int_precision else value
        return value

    # io in text files

    def save_cip(self, filename=None):
        """
        save model into a cip file

        :param (str or None) filename: the filename, by default the name of the model,
                                       if it does not have `*.cip` ending, it will be added
        """
        filename = filename or self.name
        filename = filename if filename.endswith(FILENAME_SUFFIX) else filename + FILENAME_SUFFIX

        self.writeProblem(filename)
        with open(filename, mode="a", encoding="utf-8") as txt_file:
            # add the information of the attribute in the last line
            txt_file.writelines(SAVE_ISING + f"{self.from_inverted_ising}")


    @classmethod
    def load_cip(cls, filename):
        """
        load model from a cip file

        :param (str) filename: the filename, if it does not have `*.cip` ending, it will be added
        :return the loaded model
        """
        filename = filename if filename.endswith(FILENAME_SUFFIX) else filename + FILENAME_SUFFIX
        scip_model = cls()

        with open(filename, mode="r", encoding="utf-8") as txt_file:
            last_line = txt_file.readlines()[-1]
            if last_line.startswith(SAVE_ISING):
                # get the value of the attribute in the added line
                from_inverted_ising = literal_eval(last_line[len(SAVE_ISING):])
            else:
                warn(WARNING_LOAD)
                from_inverted_ising = None

        scip_model.readProblem(filename)
        scip_model.from_inverted_ising = from_inverted_ising

        # parse the variables to store them for later usage
        scip_variables = scip_model.getVars()
        scip_variables = [var for var in scip_variables if var.name != OBJECTIVE_VAR]
        original_variables = {literal_eval(var.name.replace(VAR_PREFIX, "")): var for var in scip_variables}
        scip_model.data = original_variables
        return scip_model



# Copyright 2022 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for polynomial variables utility functions """

from numbers import Integral


NOT_ITERABLE = "is not iterable"

ERROR_TYPE     = "Variable '{}' has invalid type '{}'"
ERROR_FORMAT   = "Tuple variable '{}' is not formatted correctly, should only contain ints and strings"
ERROR_EXPECTED = "Expected variable type '{}', but got '{}'"


def get_common_type(variables):
    """
    get the type of the variables in the iterable and check all variables against this type
    (use get_type on single variable instead if no check is needed)

    :param (list or set or tuple) variables: the set of variables to be checked
    :return: the common type of the variables
    """
    if not variables:
        return None
    try:
        var_iter = iter(variables)
        some_var = next(var_iter)
    except TypeError as te:
        assert te.args[0].endswith(NOT_ITERABLE)
        some_var = variables
        var_iter = []
    var_type = get_type(some_var)
    for var in var_iter:
        check_type_against(var, var_type)
    return var_type

def get_type(variable):
    """
    get and check the type of the variable

    :param variable: the variable to be checked
    :return: the type of the variable
    """
    var_type = type(variable)
    if issubclass(var_type, Integral) and not issubclass(var_type, bool):
        var_type = Integral
    if not is_valid_variable_type(var_type):
        raise TypeError(ERROR_TYPE.format(variable, var_type.__name__))
    if issubclass(var_type, tuple) and not has_valid_tuple_format(variable):
        raise ValueError(ERROR_FORMAT.format(variable))
    if issubclass(var_type, tuple):
        var_type = tuple
    return var_type

def is_valid_variable_type(var_type):
    """
    check if variable type corresponds to our convention: either int, str, tuple or None

    :param (type) var_type: the variable type to be checked
    :return: True, if the type is valid
    """
    return issubclass(var_type, Integral) and not issubclass(var_type, bool) \
           or issubclass(var_type, tuple) or var_type in (str, None)

def has_valid_tuple_format(variable):
    """
    check whether the variable has a valid format, i.e., the tuple only consists of strings and ints

    :param variable: the variable to be checked
    :return: True, if the variable has a valid format
    """
    return isinstance(variable, tuple) and all(isinstance(v, (str, Integral)) for v in variable)

def check_type_against(variable, expected_var_type):
    """
    check the actual type of the variable against the expected variable type

    :param variable: variable object whose type is checked
    :param (type) expected_var_type: the previously determined type the variable should have
    """
    var_type = get_type(variable)
    if expected_var_type and var_type != expected_var_type:
        raise TypeError(ERROR_EXPECTED.format(expected_var_type.__name__, var_type.__name__))

def are_consecutive(variables):
    """
    check if the variables tuples do only contain integers and all appearing variables are consecutive starting with 0

    :param (list or set) variables: the iterable of variables to be checked
    :return: True, if the variables are consecutive
    """
    if variables:
        try:
            return min(variables) == 0 and len(set(variables)) == max(variables) + 1
        except TypeError:
            return False
    return True

def to_string(variable, poly_type=None, placeholder=False):
    """
    get a nice string representation of the given variable

    :param (tuple or Integral or str) variable: the variable to be formatted
    :param (type or str) poly_type: the type of the polynomial, for Ising problems the standard letter is 's',
                                    while for binary it is 'x'
    :param (bool) placeholder: the indices will be replaced with the placeholder '*'
    :param poly_type: type of the polynomial, for Ising problems the standard letter is 's', while for binary 'x'
    :param placeholder: the indices will be replaced with the placeholder '*'
    :return: the nice string representation
    """
    if isinstance(poly_type, type):
        poly_type = poly_type.__name__
    var_letter = "s" if poly_type == "PolyIsing" else "x"
    if isinstance(variable, tuple):
        var_str, start = (f"{variable[0]}_", 1) if isinstance(variable[0], str) else (f"{var_letter}_", 0)
        return var_str + "_".join(str(v) if not placeholder else "*" for v in variable[start:])
    if isinstance(variable, Integral):
        return f"{var_letter}{variable if not placeholder else '*'}"
    return str(variable)

def to_domain_string(variable, poly_type=None):
    """
    get a nice string representation of the domain of the variable

    :param (tuple or Integral or str) variable: the variable to be formatted
    :param (type or str) poly_type: the type of the polynomial, for Ising problems the standard letter is 's',
                                    while for binary it is 'x'
    :return: the nice string representation
    """
    if isinstance(poly_type, type):
        poly_type = poly_type.__name__
    domain = "{-1, 1}" if poly_type == "PolyIsing" else "{0, 1}"
    return to_string(variable, poly_type, True) + " in " + domain

def replace_strs_by_ints(tuples):
    """
    replace the strings with ints in a list of tuples themselves containing either ints or strings,
    as strings in tuples cannot be stored easily therefore they need to be replaced with an int,
    they will be enumerated beginning with the largest contained int +1

    :param (list[tuple]) tuples: list of tuples themselves containing either ints or strings
    :return: list of tuples only containing ints and a mapping from the new ints to the old strings
    """
    all_parts = set(p for var in tuples for p in var)
    str_parts = sorted(set(p for p in all_parts if isinstance(p, str)))

    # if there are strings inside we replace them by numbers larger than the maximum number appearing
    if str_parts:
        max_int = max((p for p in all_parts if isinstance(p, Integral)), default=-1)
        int_to_str = dict(zip(range(max_int + 1, max_int + 1 + len(str_parts)), str_parts))
        str_to_int = {s: i for i, s in int_to_str.items()}
        tuples = replace_in_tuples(tuples, str_to_int)
        return tuples, int_to_str
    return tuples, None

def replace_in_tuples(tuples, replacement):
    """
    replace the elements of all tuples

    :param (list[tuple]) tuples: the list of tuples
    :param (list or dict or function) replacement: dictionary or function mapping from old variable to new one
    :return: the new tuples with replaced elements
    """
    return [replace_in_tuple(t, replacement) for t in tuples]

def replace_in_tuple(var_tuple, replacement):
    """
    create a new tuple with elements from the original tuple replaced by new values defined by replacement

    :param (tuple) var_tuple: the tuple in which the variables will be replaced
    :param (list or dict or function) replacement: dictionary or function mapping from old variable to new one
    :return: the new variable tuple
    """
    if callable(replacement):
        return tuple(map(replacement, var_tuple))
    return tuple(replacement[k] if k in replacement else k for k in var_tuple)



# Copyright 2022 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for helper functions for logging """

import os
import logging
from types import MethodType


LOG_FORMAT = "%(asctime)s %(name)-12s %(levelname)-8s %(message)s"


def get_logger(logger_name,
               log_level=logging.INFO,
               log_level_console=logging.INFO,
               log_level_file=logging.NOTSET,
               log_file=None,
               log_folder=None):
    """
     get the default logger

    :param (str) logger_name: name of the logger, if already present, overwrite
    :param (logging.level) log_level: threshold log level
    :param (logging.level) log_level_console: log level for console logging, use logging.NOTSET to disable
    :param (logging.level) log_level_file: log level for file logging, use logging.NOTSET to disable
    :param (str) log_file: path to log file, by default logger_name with suffix .log
                           only applicable if log_level_file is not logging.NOTSET,
    :param (str) log_folder: if given, the full log file path is created by concatenating log_folder and log_file,
                             create folder if not existent, only applicable if log_level_file is not logging.NOTSET
    :return: the logger
    """
    logger = logging.getLogger(logger_name)
    # add close_all_handlers function to logger
    logger.close_all_handlers = MethodType(close_all_handlers, logger)
    # close all handlers (in case logger was used before)
    logger.close_all_handlers()

    logger.setLevel(log_level)
    formatter = logging.Formatter(LOG_FORMAT)

    # configure console handler
    if log_level_console != logging.NOTSET:
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(formatter)
        console_handler.setLevel(log_level_console)
        logger.addHandler(console_handler)

    # configure file handler
    if log_level_file != logging.NOTSET:
        # configure log file
        # set log file to logger_name if none is given
        if log_file is None:
            log_file = logger_name + ".log"
        # create logging folder if log_folder option is given
        if log_folder:
            if not os.path.exists(log_folder):
                os.makedirs(log_folder)
            # prepend log_folder to log_file
            log_file = log_folder + "/" + log_file

        # configure handler
        file_handler = logging.FileHandler(log_file, mode="w")
        file_handler.setFormatter(formatter)
        file_handler.setLevel(log_level_file)
        logger.addHandler(file_handler)

    return logger

def close_all_handlers(logger):
    """
    close all handlers of logger

    :param (logging.Logger) logger: logger to close all handlers
    """
    handlers = list(logger.handlers)
    for handler in handlers:
        handler.close()
        logger.removeHandler(handler)



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for testing scip_model """

import os
import pytest

from quark import Objective, Polynomial, PolyBinary, PolyIsing, ScipModel, ConstrainedObjective, ConstraintBinary
from quark.scip_model import QUADRATIC_OBJECTIVE, OPTIMAL


X, B, G, R = "x", "blue", "green", "red"

TEST_DIR = os.path.dirname(os.path.abspath(__file__)) + "/"
FILENAME_TMP = TEST_DIR + "scip_model.cip"
FILENAME_FIX = TEST_DIR + "testdata/scip_model.cip"

POLY = Polynomial.read_from_string("- 2 x0 x4 - 3 x0 x5 + x0 x6 + 3 x0 x7 + 4 x1 x4 "
                                   "+ 3 x1 x6 + 4 x0 - 3 x1 - 4 x0 - 5 x2 - 4 x5 + 16")
EXP_VARIABLES = [(X, 1, B), (X, 1, G), (X, 1, R), (X, 2, B), (X, 2, G), (X, 2, R), (X, 3, B), (X, 3, G), (X, 3, R),
                 (X, 4, B), (X, 4, G), (X, 4, R), (X, 5, B), (X, 5, G), (X, 5, R)]

@pytest.fixture(name="scip_model")
def fixture_scip_model(objective):
    """ provide scip model for testing """
    yield ScipModel.get_from_objective(objective)

def test_get_from_objective(scip_model, solutions):
    """ test solution of scip model created from objective """
    constraint = scip_model.getConss()[0]
    assert list(scip_model.data.keys()) == EXP_VARIABLES
    assert constraint.name == QUADRATIC_OBJECTIVE
    assert constraint.isNonlinear()

    solution_solve = scip_model.solve()
    assert scip_model.getObjVal() == 0
    assert scip_model.getVal(scip_model.data[(X, 1, B)]) == 1
    assert scip_model.getVal(scip_model.data[(X, 2, B)]) == 0
    assert dict(solution_solve) in solutions
    assert solution_solve.objective_value == solutions[0].objective_value

def test_get_from_constrained_objective(constrained_objective, solutions):
    """ test solution of scip model created from objective """
    scip_model = ScipModel.get_from_constrained_objective(constrained_objective)
    constraints = scip_model.getConss()
    assert list(scip_model.data.keys()) == EXP_VARIABLES
    assert len(constraints) == 6
    assert constraints[0].name == QUADRATIC_OBJECTIVE
    assert constraints[0].isNonlinear()

    for node in range(1, 6):
        assert constraints[node].name == f"one_color_per_node_{node}"
        assert constraints[node].isLinear()

    solution_solve = scip_model.solve()
    assert scip_model.getObjVal() == 0
    assert dict(solution_solve) in solutions
    assert solution_solve.objective_value == solutions[0].objective_value

def test_get_from_manual_constrained_objective():
    """ test solution of scip model created from objective """
    objective_poly = PolyBinary({(0,): 1, (1,): -1})
    with pytest.warns(UserWarning, match=r"Useless constraint '0 <= \+1 x0 <= 1'"):
        constraint = ConstraintBinary(PolyBinary({(0,): 1}), 0, 1)
    constrained_objective = ConstrainedObjective(objective_poly=objective_poly, constraints={"useless": constraint})
    scip_model = ScipModel.get_from_constrained_objective(constrained_objective)
    solution = scip_model.solve(**{"limits/nodes": 100})
    assert scip_model.getObjVal() == -1
    assert solution.objective_value == -1
    assert solution == {0: 0, 1: 1}

def test_get_from_manual_constrained_objective_infeasible():
    """ test solution of scip model created from objective """
    objective_poly = PolyBinary({(0,): 1, (1,): -1})
    constraint = ConstraintBinary(PolyBinary({(0,): 1}), 2, 3, _check=False)
    constrained_objective = ConstrainedObjective(objective_poly=objective_poly, constraints={"infeasible": constraint})
    scip_model = ScipModel.get_from_constrained_objective(constrained_objective)
    solution = scip_model.solve()
    assert not solution
    assert solution.solving_status == "infeasible"

    with pytest.raises(ValueError, match="Cannot get value for variable '2'"):
        scip_model.get_value(2)
    scip_model.data.update({2: 1})
    assert scip_model.get_value(2) == 1

def test_solve():
    """ test solution of different objectives """
    _assert_solve(Objective(PolyBinary(POLY), "binary"), 1.0, {0: 1, 1: 1, 2: 1, 4: 0, 5: 1, 6: 0, 7: 0})
    _assert_solve(Objective(PolyIsing(POLY),  "ising"), -8.0, {0: 1, 1: 1, 2: 1, 4: -1, 5: 1, 6: -1, 7: -1})
    _assert_solve(Objective(PolyIsing(POLY).invert(), "ising_inverted"), -8.0,
                  {0: -1, 1: -1, 2: -1, 4: 1, 5: -1, 6: 1, 7: 1})

def _assert_solve(objective, exp_objective_value, exp_solution):
    model = ScipModel.get_from_objective(objective)
    solution = model.solve()
    assert solution.solving_status == OPTIMAL
    assert solution.objective_value == exp_objective_value#
    assert solution == exp_solution

    score_evaluated = objective.polynomial.evaluate(solution)
    assert score_evaluated == exp_objective_value

def test_load_cip(scip_model):
    """ test loading of SCIP model """
    with pytest.warns(UserWarning, match="Cannot retrieve value for 'from_inverted_ising', will use 'None'"):
        loaded = ScipModel.load_cip(FILENAME_FIX)
    scip_model.from_inverted_ising = None
    _check_model(loaded, scip_model)

def test_io_cip(scip_model):
    """ test IO round trip """
    scip_model.save_cip(FILENAME_TMP)
    loaded = scip_model.load_cip(FILENAME_TMP)
    _check_model(loaded, scip_model)
    os.remove(FILENAME_TMP)

def _check_model(loaded, scip_model):
    assert loaded.getProbName() == scip_model.getProbName()
    assert loaded.data.keys() == scip_model.data.keys()
    assert loaded.from_inverted_ising == scip_model.from_inverted_ising

    loaded_constraint = loaded.getConss()[0]
    assert loaded_constraint.isNonlinear()
    assert loaded_constraint.name == QUADRATIC_OBJECTIVE

    loaded_objective = loaded.getObjective()
    objective = scip_model.getObjective()
    loaded_vars = [term[0].name for term in loaded_objective.terms]
    objective_vars = [term[0].name for term in objective.terms]
    assert set(loaded_vars) == set(objective_vars)



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for testing the IO of the Polynomial """

import os
import pytest

from quark import Polynomial
from quark.io import hdf5, txt


TEST_DIR = os.path.dirname(os.path.abspath(__file__)) + "/../"
FILE_H5 = "polynomial.h5"
FILE_TXT = "polynomial.txt"
FILENAME_H5_TMP = TEST_DIR + FILE_H5
FILENAME_TXT_TMP = TEST_DIR + FILE_TXT
FILENAME_H5_FIX = TEST_DIR + "testdata/" + FILE_H5
FILENAME_TXT_FIX = TEST_DIR + "testdata/" + FILE_TXT

POLY_FIX = Polynomial({(("x", 1),): 2.5, (("x", 2), ("y", 0, 0)): 2.0, (): 5.5})
POLYS = [Polynomial({(1,): 5, (2,): 6, (3, 1): 3, (2, 4): 3, (4,): 1, (): 3}),
         Polynomial({(2,): 1, (2, 4): 1, (3, 4): 3, (5,): 7}),
         Polynomial({(1,): 7, (2,): 2, (3,): 3, (4,): 1}),
         Polynomial({}),
         Polynomial({(): 55}),
         Polynomial({(): 55, (1,): 66}),
         Polynomial({(): 55, (1,): 66, (1, 2): 77}),
         Polynomial({(1,): 66, (1, 2): 77}),
         Polynomial({(1,): 1.5, (2,): 2.5}),
         Polynomial({((1,),): 1.0, ((1, 2),): 2.0, ((1, 2), (1, 2), (1, 3)): 2.0, (): 5.5}),
         POLY_FIX]


def test_load():
    """ test hdf5 io """
    loaded = hdf5.load(Polynomial, FILENAME_H5_FIX)
    assert loaded == POLY_FIX

    loaded = txt.load(Polynomial, FILENAME_TXT_FIX)
    assert loaded == POLY_FIX

@pytest.mark.parametrize("poly", POLYS)
def test_io(poly):
    """ test hdf5 io round trip """
    hdf5.save(poly, FILENAME_H5_TMP)
    loaded = hdf5.load(Polynomial, FILENAME_H5_TMP)
    assert loaded == poly

    txt.save(poly, FILENAME_TXT_TMP)
    loaded = txt.load(Polynomial, FILENAME_TXT_TMP)
    assert loaded == poly

    if poly.is_flat():
        str_poly = str(poly)
        loaded = Polynomial.read_from_string(str_poly)
    assert loaded == poly

    os.remove(FILENAME_H5_TMP)
    os.remove(FILENAME_TXT_TMP)

@pytest.mark.parametrize("poly", POLYS)
def test_io_methods(poly):
    """ test hdf5 io round trip """
    poly.save_hdf5(FILENAME_H5_TMP)
    loaded = Polynomial.load_hdf5(FILENAME_H5_TMP)
    assert loaded == poly

    poly.save_txt(FILENAME_TXT_TMP)
    loaded = Polynomial.load_txt(FILENAME_TXT_TMP)
    assert loaded == poly

    os.remove(FILENAME_H5_TMP)
    os.remove(FILENAME_TXT_TMP)

def test_existence_save_load():
    assert hasattr(Polynomial, "save_hdf5")
    assert hasattr(Polynomial, "load_hdf5")
    assert hasattr(Polynomial, "exists_hdf5")
    assert hasattr(Polynomial, "save_txt")
    assert hasattr(Polynomial, "load_txt")
    assert hasattr(Polynomial, "exists_txt")



# Copyright 2023 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for ExampleObjective """

from quark import Objective, PolyBinary


X = "x"


class ExampleObjective(Objective):
    """ An example implementation of an Objective for testing based on the MaxColorableSubgraph problem """

    @staticmethod
    def _get_polynomial(instance):
        """ get the objective terms from the instance data """
        # actual objective
        # counting the number of wrongly colored edges:
        # sum_[c in Colors] sum_[(n, m) in Edges] (1 * x_n_c * x_m_c)
        poly_colored_edges = PolyBinary()
        for color in instance.colors:
            poly_colored_edges += PolyBinary({((X, node1, color), (X, node2, color)): 1
                                              for node1, node2 in instance.edges})

        # additional penalty term from the constraint
        # every node should get exactly one color:
        # sum_[n in Nodes] (sum_[c in Colors] x_n_c - 1)^2
        poly_one_color_per_node = PolyBinary()
        for node in instance.nodes:
            poly = PolyBinary({((X, node, color),): 1 for color in instance.colors}) - 1
            poly_one_color_per_node += poly * poly
        # apply a default worst case scaling
        poly_one_color_per_node *= len(instance.edges)

        # return the full QUBO
        return poly_colored_edges + poly_one_color_per_node



# Copyright 2024 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for fixtures for sophisticated tests """

import pytest

from quark import PolyBinary, ConstraintBinary, ConstrainedObjective, ObjectiveTerms, Objective, Solution
from quark.testing import ExampleInstance, ExampleObjectiveTerms, ExampleObjective


X = "x"
EXP_NAME = "colored_edges1.000000e+00_one_color_per_node7.000000e+00"
B, G, R = "blue", "green", "red"


@pytest.fixture(name="instance")
def fixture_instance():
    """
    provide the ExampleInstance test object,
    built upon an example implementation inheriting from the base class
    """
    edges = [(1, 2), (2, 3), (3, 4), (4, 5), (5, 1), (1, 3), (3, 5)]
    colors = [G, B, R]
    yield ExampleInstance(edges, colors)

@pytest.fixture(name="constrained_objective")
def fixture_constrained_objective(instance):
    """
    provide ConstrainedObjective test object from instance data,
    uses ExampleConstrainedObjective implementation
    """
    yield instance.get_constrained_objective()

@pytest.fixture(name="objective_terms")
def fixture_objective_terms(instance):
    """
    provide ObjectiveTerms test object from instance data,
    uses ExampleConstrainedObjective implementation and automized ObjectiveTerms construction
    """
    yield instance.get_objective_terms()

@pytest.fixture(name="objective")
def fixture_objective(instance):
    """
    provide Objective test object from instance data,
    uses ExampleConstrainedObjective implementation and automized ObjectiveTerms and Objective construction
    """
    yield instance.get_objective(penalty_scale=7)


@pytest.fixture(name="objective_terms_direct_impl")
def fixture_objective_terms_direct_impl(instance):
    """
    provide ObjectiveTerms test object from instance data,
    uses ExampleConstrainedObjective implementation and automized ObjectiveTerms construction
    """
    yield ExampleObjectiveTerms.get_from_instance(instance)

@pytest.fixture(name="objective_direct_impl")
def fixture_objective_direct_impl(instance):
    """
    provide Objective test object from instance data,
    uses ExampleConstrainedObjective implementation and automized ObjectiveTerms and Objective construction
    """
    yield ExampleObjective.get_from_instance(instance)


EXP_OBJECTIVE_POLY = PolyBinary({((X, 1, B), (X, 2, B)): 1, ((X, 1, G), (X, 2, G)): 1, ((X, 1, R), (X, 2, R)): 1,
                                 ((X, 2, B), (X, 3, B)): 1, ((X, 2, G), (X, 3, G)): 1, ((X, 2, R), (X, 3, R)): 1,
                                 ((X, 3, B), (X, 4, B)): 1, ((X, 3, G), (X, 4, G)): 1, ((X, 3, R), (X, 4, R)): 1,
                                 ((X, 4, B), (X, 5, B)): 1, ((X, 4, G), (X, 5, G)): 1, ((X, 4, R), (X, 5, R)): 1,
                                 ((X, 5, B), (X, 1, B)): 1, ((X, 5, G), (X, 1, G)): 1, ((X, 5, R), (X, 1, R)): 1,
                                 ((X, 1, B), (X, 3, B)): 1, ((X, 1, G), (X, 3, G)): 1, ((X, 1, R), (X, 3, R)): 1,
                                 ((X, 3, B), (X, 5, B)): 1, ((X, 3, G), (X, 5, G)): 1, ((X, 3, R), (X, 5, R)): 1})
EXP_PENALTY_TERM = PolyBinary({(): 5,
                               ((X, 1, B),): -1, ((X, 1, G),): -1, ((X, 1, R),): -1,
                               ((X, 2, B),): -1, ((X, 2, G),): -1, ((X, 2, R),): -1,
                               ((X, 3, B),): -1, ((X, 3, G),): -1, ((X, 3, R),): -1,
                               ((X, 4, B),): -1, ((X, 4, G),): -1, ((X, 4, R),): -1,
                               ((X, 5, B),): -1, ((X, 5, G),): -1, ((X, 5, R),): -1,
                               ((X, 1, B), (X, 1, G)): 2, ((X, 1, B), (X, 1, R)): 2, ((X, 1, G), (X, 1, R)): 2,
                               ((X, 2, B), (X, 2, G)): 2, ((X, 2, B), (X, 2, R)): 2, ((X, 2, G), (X, 2, R)): 2,
                               ((X, 3, B), (X, 3, G)): 2, ((X, 3, B), (X, 3, R)): 2, ((X, 3, G), (X, 3, R)): 2,
                               ((X, 4, B), (X, 4, G)): 2, ((X, 4, B), (X, 4, R)): 2, ((X, 4, G), (X, 4, R)): 2,
                               ((X, 5, B), (X, 5, G)): 2, ((X, 5, B), (X, 5, R)): 2, ((X, 5, G), (X, 5, R)): 2})

@pytest.fixture(name="exp_constrained_objective")
def fixture_exp_constrained_objective():
    """ provide the fixed ObjectiveTerms test object """
    constraint_polys = [PolyBinary({((X, 1, B),): 1, ((X, 1, G),): 1, ((X, 1, R),): 1}),
                        PolyBinary({((X, 2, B),): 1, ((X, 2, G),): 1, ((X, 2, R),): 1}),
                        PolyBinary({((X, 3, B),): 1, ((X, 3, G),): 1, ((X, 3, R),): 1}),
                        PolyBinary({((X, 4, B),): 1, ((X, 4, G),): 1, ((X, 4, R),): 1}),
                        PolyBinary({((X, 5, B),): 1, ((X, 5, G),): 1, ((X, 5, R),): 1})]
    exp_constraints = {"one_color_per_node_1": ConstraintBinary(constraint_polys[0], 1, 1),
                       "one_color_per_node_2": ConstraintBinary(constraint_polys[1], 1, 1),
                       "one_color_per_node_3": ConstraintBinary(constraint_polys[2], 1, 1),
                       "one_color_per_node_4": ConstraintBinary(constraint_polys[3], 1, 1),
                       "one_color_per_node_5": ConstraintBinary(constraint_polys[4], 1, 1)}
    yield ConstrainedObjective(EXP_OBJECTIVE_POLY, exp_constraints)

@pytest.fixture(name="exp_objective_terms")
def fixture_exp_objective_terms():
    """ provide the fixed ObjectiveTerms test object """
    exp_terms = {"colored_edges": EXP_OBJECTIVE_POLY,
                 "one_color_per_node": EXP_PENALTY_TERM}
    yield ObjectiveTerms(exp_terms, ["one_color_per_node"])

@pytest.fixture(name="exp_objective")
def fixture_exp_objective():
    """ provide the fixed Objective test object """
    exp_objective_poly = EXP_OBJECTIVE_POLY + 7 * EXP_PENALTY_TERM
    yield Objective(exp_objective_poly, EXP_NAME)


VAR_ASSIGNMENTS = [{(X, 1, B): 1, (X, 1, G): 0, (X, 1, R): 0,
                    (X, 2, B): 0, (X, 2, G): 0, (X, 2, R): 1,
                    (X, 3, B): 0, (X, 3, G): 1, (X, 3, R): 0,
                    (X, 4, B): 1, (X, 4, G): 0, (X, 4, R): 0,
                    (X, 5, B): 0, (X, 5, G): 0, (X, 5, R): 1},
                   {(X, 1, B): 1, (X, 1, G): 0, (X, 1, R): 0,
                    (X, 2, B): 0, (X, 2, G): 1, (X, 2, R): 0,
                    (X, 3, B): 0, (X, 3, G): 0, (X, 3, R): 1,
                    (X, 4, B): 1, (X, 4, G): 0, (X, 4, R): 0,
                    (X, 5, B): 0, (X, 5, G): 1, (X, 5, R): 0}]

@pytest.fixture(name="solution")
def fixture_solution():
    """ provide the Solution test object """
    yield Solution(VAR_ASSIGNMENTS[0], objective_value=0.0, name='from_constrained_objective')

@pytest.fixture(name="solutions")
def fixture_solutions():
    """ provide the Solution test object """
    yield [Solution(var_assignments, objective_value=0.0, name='from_constrained_objective')
           for var_assignments in VAR_ASSIGNMENTS]



# Copyright 2022 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for testing the Embedding """

import pytest

from quark import Embedding, HardwareAdjacency


NAME = "test_embedding"
VAR_NODES_MAP = {'a': [4, 0, 24, 48],
                 'b': [5, 1, 25],
                 'c': [6, 2, 26]}
NODE_ASSIGNMENTS = {4: 1, 0: 1, 24: 1, 48: 1, 5: 0, 1: 0, 25: 1, 6: 0, 2: 0, 26: 1}

@pytest.fixture(name="embedding")
def fixture_embedding():
    """ provide the Embedding test object """
    yield Embedding(var_nodes_map=VAR_NODES_MAP, name=NAME, index=0)


def test_init(embedding):
    """ test initialization of embedding """
    assert set(embedding.nodes) == {0, 1, 2, 4, 5, 6, 48, 24, 25, 26}
    assert embedding.max_map_size == 4
    assert embedding.edges == []

def test_consistency():
    """ test correctly thrown errors """
    with pytest.raises(ValueError, match="Either provide mapping from variables to nodes or to edges"):
        Embedding(name="nothing")
    with pytest.raises(ValueError, match="The var_edges_map cannot contain different variables than var_nodes_map"):
        Embedding(var_nodes_map={0: [0], 1: [1]}, var_edges_map={0: [(0, 2)]}, name="wrong_nodes")

    embedding = Embedding(var_nodes_map={0: [0], 1: [1]}, name="no_maps")
    with pytest.raises(ValueError, match="Cannot judge validity without mapping from couplings to edges"):
        embedding.are_all_couplings_mapped([(0, 1)])
    with pytest.raises(ValueError, match="Cannot check connectivity without mapping from variables to edges"):
        embedding.are_all_subgraphs_connected()

    with pytest.raises(ValueError, match="Each variable should map to at least one node"):
        Embedding(var_nodes_map={0: [0, 4], 1: [], 2: [2, 6]}, var_edges_map={0: [(0, 4)], 1: [], 2: [(2, 6)]})

def test_repr(embedding):
    """ test string representation """
    exp_str = "Embedding({'a': [4, 0, 24, 48], 'b': [5, 1, 25], 'c': [6, 2, 26]}, 'test_embedding', 0)"
    assert repr(embedding) == exp_str

def test_str(embedding):
    """ test string representation """
    exp_str = "a \t-> [4, 0, 24, 48]\n" \
              "b \t-> [5, 1, 25]\n" \
              "c \t-> [6, 2, 26]"
    assert str(embedding) == exp_str

def test_get_from_hwa():
    """ test construction of embedding with a given HWA """
    edges = [(0, 1), (1, 2), (0, 3), (1, 4), (2, 5), (3, 4), (4, 5), (6, 3), (7, 4), (8, 5), (6, 7), (7, 8)]
    hwa = HardwareAdjacency(edges, "test_hwa")  # 3x3 grid example
    var_nodes_map = {"a": [0, 1], "b": [3, 4], "c": [2, 5]}
    embedding = Embedding.get_from_hwa(var_nodes_map, hwa)
    couplings = [("a", "b"), ("a", "c"), ("c", "b")]

    exp_embedding = Embedding(var_nodes_map={'a': [0, 1], 'b': [3, 4], 'c': [2, 5]},
                              var_edges_map={'a': [(0, 1)], 'b': [(3, 4)], 'c': [(2, 5)]},
                              coupling_edges_map={('a', 'b'): [(0, 3), (1, 4)],
                                                  ('a', 'c'): [(1, 2)],
                                                  ('b', 'c'): [(4, 5)]},
                              name='test_hwa')
    not_exp_embedding = Embedding(var_nodes_map={'a': [0, 1], 'b': [3, 4], 'c': [2, 5]},
                                  var_edges_map={'a': [(0, 1)], 'b': [(3, 4)], 'c': [(2, 5)]},
                                  coupling_edges_map={('a', 'b'): [(0, 3)],
                                                      ('a', 'c'): [(1, 2)],
                                                      ('b', 'c'): [(4, 5)]},
                                  name='test_hwa')
    assert embedding == exp_embedding
    assert embedding != not_exp_embedding
    assert embedding != {'a': [0, 1], 'b': [3, 4], 'c': [2, 5]}

    assert embedding.is_valid(couplings)
    assert set(embedding.edges) > {(0, 1), (0, 3), (1, 2)}
    assert embedding.name == hwa.name

    embedding_only_edges = Embedding.get_from_hwa(var_nodes_map, edges, name=hwa.name)
    assert embedding == embedding_only_edges

def test_checks():
    """ test the different validity checks """
    embedding = Embedding(var_nodes_map={0: [0, 4], 1: [1, 6], 2: [2, 6]})
    assert not embedding.are_all_node_sets_disjoint()
    assert not embedding.is_valid()

    embedding = Embedding(var_nodes_map={0: [0, 4], 1: [1, 5], 2: [2, 6]},
                          var_edges_map={0: [(0, 4)], 1: [], 2: [(2, 6)]})
    assert not embedding.are_all_subgraphs_connected()
    assert not embedding.is_valid()

    embedding = Embedding(var_nodes_map={0: [0, 4], 1: [1], 2: [2, 6]},
                          var_edges_map={0: [(0, 4)], 1: [], 2: [(2, 6)]})
    assert embedding.are_all_subgraphs_connected()
    assert embedding.is_valid()

    embedding = Embedding(coupling_edges_map={(0, 1): [(0, 5)]},
                          var_edges_map={0: [(0, 4)], 1: [(1, 5)], 2: [(2, 6)]})
    assert not embedding.are_all_couplings_mapped([(0, 2)])
    assert embedding.is_valid()
    assert not embedding.is_valid([(0, 2)])

    embedding = Embedding(coupling_edges_map={(0, 1): [(0, 5)]},
                          var_edges_map={0: [(0, 4)], 1: [(1, 7)], 2: [(2, 6)]})
    assert not embedding.are_all_edges_valid()
    assert not embedding.is_valid()

def test_get_single_var_scores(embedding):
    """ test scores of variable assignments for de-embedding """
    assert embedding.get_var_value_scores(NODE_ASSIGNMENTS) == [("a", [(1, 1.0)]),
                                                                ("b", [(0, 2/3), (1, 1/3)]),
                                                                ('c', [(0, 2/3), (1, 1/3)])]

def test_vars_with_broken_embedding(embedding):
    """ test broken embeddings """
    assert embedding.get_vars_with_broken_embedding(NODE_ASSIGNMENTS) == ['b', 'c']

def test_de_embed(embedding):
    """ test de-embedding """
    assert embedding.de_embed(NODE_ASSIGNMENTS) == {'a': 1, 'b': 0, 'c': 0}
    assert embedding.de_embed(NODE_ASSIGNMENTS, return_score=True) == ({'a' : 1, 'b': 0, 'c': 0}, 4/9)

    assert not embedding.de_embed(NODE_ASSIGNMENTS, single_var_threshold=0.8)
    assert embedding.de_embed(NODE_ASSIGNMENTS, single_var_threshold=0.8, return_score=True) == ([], 4/9)

    assert embedding.de_embed(NODE_ASSIGNMENTS, single_var_threshold=0.6) == [{'a': 1, 'b': 0, 'c': 0}]
    assert embedding.de_embed(NODE_ASSIGNMENTS, single_var_threshold=0.6, return_score=True) \
                                                                          == [({'a': 1, 'b': 0, 'c': 0}, 4/9)]

    assert not embedding.de_embed(NODE_ASSIGNMENTS, total_threshold=0.8)
    assert embedding.de_embed(NODE_ASSIGNMENTS, total_threshold=0.8, return_score=True) == ([], 4/9)

    assert embedding.de_embed(NODE_ASSIGNMENTS, total_threshold=0.2) == [{'a': 1, 'b': 0, 'c': 0},
                                                                         {'a': 1, 'b': 1, 'c': 0},
                                                                         {'a': 1, 'b': 0, 'c': 1}]
    assert embedding.de_embed(NODE_ASSIGNMENTS, total_threshold=0.3) == [{'a': 1, 'b': 0, 'c': 0}]
    assert embedding.de_embed(NODE_ASSIGNMENTS, total_threshold=0.2, return_score=True) \
                                                                     == [({'a': 1, 'b': 0, 'c': 0}, 4/9),
                                                                         ({'a': 1, 'b': 1, 'c': 0}, 2/9),
                                                                         ({'a': 1, 'b': 0, 'c': 1}, 2/9)]
    assert embedding.de_embed(NODE_ASSIGNMENTS, total_threshold=0.3, return_score=True) \
                                                                     == [({'a': 1, 'b': 0, 'c': 0}, 4/9)]
